import csv
import json
import sqlite3
import hashlib
from .beaker import Beaker
from structlog import get_logger

log = get_logger()


class Pour:
    def __init__(self, from_beaker: Beaker, to_beaker: Beaker, transform: callable):
        self.from_beaker = from_beaker
        self.to_beaker = to_beaker
        self.transform = transform


def get_sha512(filename: str) -> str:
    with open(filename, "rb") as file:
        return hashlib.sha512(file.read()).hexdigest()


class Recipe:
    def __init__(self, name: str):
        self.name = name
        self.pours = []
        self.db = sqlite3.connect("beakers.db")
        self.cursor = self.db.cursor()
        self.cursor.row_factory = sqlite3.Row
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS _metadata (table_name TEXT PRIMARY KEY, data JSON)"
        )

    def __repr__(self) -> str:
        return f"Recipe({self.name})"

    def get_metadata(self, table_name) -> dict:
        self.cursor.execute(
            "SELECT data FROM _metadata WHERE table_name = ?",
            (table_name,),
        )
        try:
            data = self.cursor.fetchone()["data"]
            log.debug("get_metadata", table_name=table_name, data=data)
            return json.loads(data)
        except TypeError:
            log.debug("get_metadata", table_name=table_name, data={})
            return {}

    def save_metadata(self, table_name: str, data: dict) -> None:
        data_json = json.dumps(data)
        log.info("save_metadata", table_name=table_name, data=data_json)
        # sqlite upsert
        self.cursor.execute(
            "INSERT INTO _metadata (table_name, data) VALUES (?, ?) ON CONFLICT(table_name) DO UPDATE SET data = ?",
            (table_name, data_json, data_json),
        )
        self.db.commit()

    def add_pour(self, from_beaker: Beaker, to_beaker: Beaker, transform: callable):
        pour = Pour(from_beaker, to_beaker, transform)
        self.pours.append(pour)

    def csv_to_beaker(self, filename: str, beaker: Beaker) -> None:
        beaker.connect_to_db(self.db)
        lg = log.bind(beaker=beaker, filename=filename)
        # three cases: empty, match, mismatch
        # case 1: empty
        if len(beaker) == 0:
            with open(filename, "r") as file:
                reader = csv.DictReader(file)
                added = 0
                for row in reader:
                    beaker.add_item(row)
                    added += 1
            lg.info("from_csv", case="empty", added=added)
            meta = self.get_metadata(beaker.table_name)
            meta["sha512"] = get_sha512(filename)
            self.save_metadata(beaker.table_name, meta)
        else:
            old_sha = self.get_metadata(beaker.table_name).get("sha512")
            new_sha = get_sha512(filename)
            if old_sha != new_sha:
                # case 3: mismatch
                lg.info("from_csv", case="mismatch", old_sha=old_sha, new_sha=new_sha)
                raise Exception("sha512 mismatch")
            else:
                # case 2: match
                lg.info("from_csv", case="match")
        return beaker

    def add_pour(self, from_beaker: Beaker, to_beaker: Beaker, transform: callable):
        pour = Pour(from_beaker, to_beaker, transform)
        self.pours.append(pour)

    def run_linearly(self):
        log.info("recipe", recipe=self)
        for pour in self.pours:
            pour.from_beaker.connect_to_db(self.db)
            pour.to_beaker.connect_to_db(self.db)

            log.info(
                "pour",
                from_beaker=pour.from_beaker,
                to_beaker=pour.to_beaker,
                to_pour=len(pour.from_beaker),
            )
            for id, item in pour.from_beaker:
                log.info("pour_item", id=id, item=item)
                transformed = pour.transform(item)
                pour.to_beaker.add_item(transformed, pour.from_beaker.table_name, id)
